import { ExplicitParams } from "./config-loader";
/**
 * Installs a custom module load function that can adhere to paths in tsconfig.
 * Returns a function to undo paths registration.
 */
export declare function register(explicitParams: ExplicitParams): () => void;
